<?php
/**
 *
 * Copyright © 2016 Magento. All rights reserved.
 * See COPYING.txt for license details.
 */

namespace FiloBlu\OrderAdditionalData\Model\Plugin;

use Exception;
use FiloBlu\OrderAdditionalData\Model\AdditionalDataFactory;
use FiloBlu\OrderAdditionalData\Model\ItemDataFactory;
use FiloBlu\OrderAdditionalData\Model\LengowOrderFactory;
use FiloBlu\OrderAdditionalData\Model\OrderDataFactory;
use FiloBlu\OrderAdditionalData\Model\ProductOptionFactory;
use Magento\Catalog\Api\ProductCustomOptionRepositoryInterface;
use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Framework\App\Helper\Context;
use Magento\Framework\App\ObjectManager;
use Magento\Framework\App\ResourceConnection;
use Magento\Framework\Exception\NoSuchEntityException;
use Magento\Framework\Module\Manager;
use Magento\Framework\ObjectManagerInterface;
use Magento\Framework\Serialize\SerializerInterface;
use Magento\GiftMessage\Api\Data\MessageInterface;
use Magento\GiftWrapping\Model\WrappingFactory;
use Magento\Sales\Api\Data\OrderExtension;
use Magento\Sales\Api\Data\OrderExtensionFactory;
use Magento\Sales\Api\Data\OrderInterface;
use Magento\Sales\Api\Data\OrderItemExtension;
use Magento\Sales\Api\Data\OrderItemExtensionFactory;
use Magento\Sales\Api\Data\OrderItemInterface;
use Magento\Sales\Api\OrderRepositoryInterface;
use Magento\Sales\Model\ResourceModel\Order\Collection;
use Magento\Store\Model\ScopeInterface;
use Magento\Store\Model\StoreManagerInterface;

/**
 * Class OrderGet
 * Load Additional Data when getting Order by OrderRepository
 * @package FiloBlu\OrderAdditionalData\Model\Plugin
 */
class OrderGet
{
    const XML_PATH_ALLOWED_FOR_ORDER = 'sales/gift_options/wrapping_allow_order';

    /** @var OrderDataFactory */
    protected $orderDataFactory;

    /** @var ItemDataFactory */
    protected $itemDataFactory;

    /** @var OrderExtensionFactory */
    protected $orderExtensionFactory;

    /** @var OrderItemExtensionFactory */
    protected $orderItemExtensionFactory;

    /** @var ResourceConnection */
    protected $resourceConnection;

    /**
     * @var ScopeConfigInterface
     */
    protected $scopeConfig;

    /**
     * @var StoreManagerInterface
     */
    protected $_storeManager;

    /**
     * @var AdditionalDataFactory
     */
    protected $_additionalDataFactory;

    /**
     * @var AdditionalDataFactory
     */
    protected $_productOptionFactory;

    /**
     * @var ProductCustomOptionRepositoryInterface
     */
    protected $_customOptionRepository;
    /**
     * @var LengowOrderFactory
     */
    protected $_lengowOrderFactory;
    /**
     * @var Manager
     */
    private $moduleManager;
    /**
     * @var \Magento\Framework\ObjectManagerInterface
     */
    private  $objectManager;

    /**
     * @param \Magento\Framework\App\Helper\Context $context
     * @param \Magento\Store\Model\StoreManagerInterface $storeManager
     * @param \FiloBlu\OrderAdditionalData\Model\OrderDataFactory $orderDataFactory
     * @param \FiloBlu\OrderAdditionalData\Model\ItemDataFactory $itemDataFactory
     * @param \Magento\Sales\Api\Data\OrderExtensionFactory $orderExtensionFactory
     * @param \Magento\Sales\Api\Data\OrderItemExtensionFactory $orderItemExtensionFactory
     * @param \FiloBlu\OrderAdditionalData\Model\AdditionalDataFactory $additionalDataFactory
     * @param \FiloBlu\OrderAdditionalData\Model\ProductOptionFactory $productOptionFactory
     * @param \FiloBlu\OrderAdditionalData\Model\LengowOrderFactory $lengowOrderFactory
     * @param \Magento\Catalog\Api\ProductCustomOptionRepositoryInterface $customOptionRepository
     * @param \Magento\Framework\App\ResourceConnection $resourceConnection
     * @param \Magento\Framework\ObjectManagerInterface $objectManager
     */
    public function __construct(
        Context $context,
        StoreManagerInterface $storeManager,
        OrderDataFactory $orderDataFactory,
        ItemDataFactory $itemDataFactory,
        OrderExtensionFactory $orderExtensionFactory,
        OrderItemExtensionFactory $orderItemExtensionFactory,
        AdditionalDataFactory $additionalDataFactory,
        ProductOptionFactory $productOptionFactory,
        LengowOrderFactory $lengowOrderFactory,
        ProductCustomOptionRepositoryInterface $customOptionRepository,
        ResourceConnection $resourceConnection,
        ObjectManagerInterface $objectManager
    ) {
        $this->orderDataFactory = $orderDataFactory;
        $this->itemDataFactory = $itemDataFactory;
        $this->orderExtensionFactory = $orderExtensionFactory;
        $this->orderItemExtensionFactory = $orderItemExtensionFactory;
        $this->scopeConfig = $context->getScopeConfig();
        $this->_storeManager = $storeManager;
        $this->_additionalDataFactory = $additionalDataFactory;
        $this->_productOptionFactory = $productOptionFactory;
        $this->_lengowOrderFactory = $lengowOrderFactory;
        $this->_customOptionRepository = $customOptionRepository;
        $this->resourceConnection = $resourceConnection;
        $this->moduleManager = $context->getModuleManager();
        $this->objectManager = $objectManager;
    }

    /**
     * @param OrderRepositoryInterface $subject
     * @param Collection $resultOrder
     * @return Collection
     */
    public function afterGetList(
        OrderRepositoryInterface $subject,
        Collection $resultOrder
    ) {
        /** @var  $order */
        foreach ($resultOrder->getItems() as $order) {
            $this->afterGet($subject, $order);
        }
        return $resultOrder;
    }

    /**
     * Add Data to order
     *
     * @param OrderRepositoryInterface $subject
     * @param OrderInterface $resultOrder
     * @return OrderInterface
     */
    public function afterGet(
        OrderRepositoryInterface $subject,
        OrderInterface $resultOrder
    ) {
        $resultOrder = $this->getOrderAdditionalData($resultOrder);
        $resultOrder = $this->getOrderItemAdditionalData($resultOrder);
        $resultOrder = $this->getMspData($resultOrder);
        $resultOrder = $this->getLengowData($resultOrder);
        return $resultOrder;
    }

    /**
     * Get additional data for order
     *
     * @param OrderInterface $order
     * @return OrderInterface
     */
    protected function getOrderAdditionalData(OrderInterface $order)
    {
        $extensionAttributes = $order->getExtensionAttributes();
        if ($extensionAttributes && $extensionAttributes->getFbOad()) {
            return $order;
        }
        //$additionalData=$this->objectFactory->create();
        $additionalData = [];
        try {
            /** @var \Magento\AdditionalData\Api\Data\MessageInterface $AdditionalData */
            $orderDataObject = $this->orderDataFactory->create()->load($order->getEntityId(), 'order_id');

            if ($orderDataObject && $orderDataObject->getID() > 0 && $orderDataObject->getAdditionalData() != null) {
                $additionalData = $this->unserialize($orderDataObject->getAdditionalData());
            }
        } catch (NoSuchEntityException $e) {
            return $order;
        }

        /** @var OrderExtension $orderExtension */
        $orderExtension = $extensionAttributes ?: $this->orderExtensionFactory->create();
        $additionalDataObject = $this->_additionalDataFactory->create();
        foreach ($additionalData as $field => $value) {
            $additionalDataObject->setCustomAttribute($field, $value);
        }

        $order_fields_config = $this->scopeConfig->getValue('oad/general/order_data_fields', ScopeInterface::SCOPE_STORE);

        $order_fields = null;
        if($order_fields_config) {
            $order_fields = $this->unserialize($order_fields_config);
        }

        if ($order_fields) {
            foreach ($order_fields as $order_field) {
                if ($order_field['object_type'] === 'order') {
                    $additionalDataObject->setCustomAttribute($order_field['fieldname'], $order->getData($order_field['object_field']));
                }
            }
        }
        $orderExtension->setFbOad($additionalDataObject);
        $orderExtension->setData('fbnetsuite_base_theoric_shipping_amount', $order->getData('fbnetsuite_base_theoric_shipping_amount'));
        $orderExtension->setData('fbnetsuite_theoric_shipping_amount', $order->getData('fbnetsuite_theoric_shipping_amount'));
        $orderExtension->setData('shop_code', $this->getShopCode($order->getPickupInStoreId()));
        $order->setExtensionAttributes($orderExtension);

        $this->addGiftWrapping($order);

        return $order;
    }

    /**
     * Get additional data for order items
     *
     * @param OrderInterface $order
     * @return OrderInterface
     */
    protected function getOrderItemAdditionalData(OrderInterface $order)
    {
        $orderItems = $order->getItems();
        if (null !== $orderItems) {
            $item_fields = $this->scopeConfig->getValue('oad/general/item_data_fields', ScopeInterface::SCOPE_STORE);
            if ($item_fields) {
                $item_fields = $this->unserialize($item_fields);
                foreach ($item_fields as $key => $item_field) {
                    if ($item_field['object_type'] != 'order_item') {
                        unset($item_fields[$key]);
                    }
                }
            }

            /** @var OrderItemInterface $orderItem */
            foreach ($orderItems as $orderItem) {
                $extensionAttributes = $orderItem->getExtensionAttributes();
                if ($extensionAttributes && $extensionAttributes->getFbOad()) {
                    continue;
                }

                try {
                    /* @var MessageInterface $giftMessage */
                    $itemDataObject = $this->itemDataFactory->create()->load($orderItem->getItemId(), 'item_id');
                    if (!is_null($itemDataObject->getAdditionalData())) {
                        $additionalData = $this->unserialize($itemDataObject->getAdditionalData());
                    } else {
                        return $order;
                    }
                } catch (NoSuchEntityException $e) {
                    continue;
                }
                if ($item_fields) {
                    foreach ($item_fields as $item_field) {
                        $additionalData[$item_field['fieldname']] = $orderItem->getData($item_field['object_field']);
                    }
                }
                /** @var OrderItemExtension $orderItemExtension */
                $orderItemExtension = $extensionAttributes ?: $this->orderItemExtensionFactory->create();
                $additionalDataObject = $this->_additionalDataFactory->create();
                foreach ($additionalData as $field => $value) {
                    $additionalDataObject->setCustomAttribute($field, $value);
                }
                $orderItemExtension->setFbOad($additionalDataObject);
                $orderItemExtension->setData('kpi_full_price', floatval($orderItem->getData('kpi_full_price')));
                if($orderItem->getProduct()){
                    $orderItemExtension->setData('base_sku', $orderItem->getProduct()->getSku());
                }
                $allProductOptions = $orderItem->getProductOptions();
                if (isset($allProductOptions['options']) && !empty($allProductOptions['options'])) {
                    $productOptions = $allProductOptions['options'];
                    $productOptionsArray = [];
                    foreach ($productOptions as $key => $product_option) {
                        if($orderItem->getProduct()){
                            $productOptionsArray[$key] = $this->_productOptionFactory->create();

                            $custom_option = $this->_customOptionRepository->get($orderItem->getProduct()->getSku(), $product_option['option_id']);
                            $default_value = $product_option['value'];
                            $option_sku = '';
                            $values = $custom_option->getValues();
                            if ($values) {
                                foreach ($values as $value) {
                                    if ($value->getOptionTypeId() == $product_option['option_value']) {
                                        $default_value = $value->getDefaultTitle();
                                        $option_sku = $value->getSku();
                                        break;
                                    }
                                }
                            }
                            $productOptionsArray[$key]->setData('code', $custom_option->getCode());
                            $productOptionsArray[$key]->setData('value_default_title', $default_value);
                            $productOptionsArray[$key]->setData('sku', $option_sku);
                            foreach ($product_option as $field => $value) {
                                $productOptionsArray[$key]->setData($field, $value);
                            }
                        }
                    }
                    $orderItemExtension->setProductOptions($productOptionsArray);
                }
                if (array_key_exists('giftcard_created_codes', $allProductOptions)) {
                    $orderItemExtension->setData('giftcard_created_codes', $allProductOptions['giftcard_created_codes']);
                }
                $orderItem->setExtensionAttributes($orderItemExtension);
            }
        }
        return $order;
    }

    /**
     * (temporary) Get MSP additional data for order
     *
     * @param OrderInterface $order
     * @return OrderInterface
     */
    protected function getMspData(OrderInterface $order)
    {
        if(!$order->getPayment()) {
            return $order;
        }
        
        $extensionAttributes = $order->getExtensionAttributes();
        if ($extensionAttributes && $extensionAttributes->getPaymentFeeAmount()) {
            return $order;
        }
        if ($order->getPayment()->getMethod() !== 'msp_cashondelivery') {
            return $order;
        }
        /** @var OrderExtension $orderExtension */
        $orderExtension = $extensionAttributes ?: $this->orderExtensionFactory->create();
        $orderExtension->setPaymentFeeAmount((float)$order->getMspCodAmount() - (float)$order->getMspCodTaxAmount());
        $orderExtension->setBasePaymentFeeAmount((float)$order->getBaseMspCodAmount() - (float)$order->getBaseMspCodTaxAmount());
        $orderExtension->setPaymentTaxAmount((float)$order->getMspCodTaxAmount());
        $orderExtension->setBasePaymentTaxAmount((float)$order->getBaseMspCodTaxAmount());
        $order->setExtensionAttributes($orderExtension);

        return $order;
    }

    /**
     * (temporary) Get Lengow Data if available
     *
     * @param OrderInterface $order
     * @return OrderInterface
     */
    protected function getLengowData(OrderInterface $order)
    {
        if(!$order->getPayment()) {
            return $order;
        }
        
        if ($order->getPayment()->getMethod() !== 'lengow') {
            return $order;
        }
        /* If lengow module is enabled the function returns the Lengow Order ID.
         * Otherwise the function is considered a magic function within a Magento\Framework\DataObject and returns null.
         * */
        $lengowOrderId = $this->_lengowOrderFactory->create()->getLengowOrderIdByOrderId($order->getId());
        if ($lengowOrderId) {
            $lengowOrder = $this->_lengowOrderFactory->create()->load($lengowOrderId);

            $extensionAttributes = $order->getExtensionAttributes();
            $orderExtension = $extensionAttributes ?: $this->orderExtensionFactory->create();

            /** @var OrderExtension $orderExtension */
            $orderExtension->setData('lengow_marketplace_name', $lengowOrder->getData('marketplace_name'));
            $orderExtension->setData('lengow_marketplace_sku', $lengowOrder->getData('marketplace_sku'));
            $extra = json_decode($lengowOrder->getData('extra'), true);
            if (is_array($extra) && array_key_exists('processing_fee', $extra)) {
                $orderExtension->setData('lengow_marketplace_processing_fee', $extra['processing_fee']);
            }
            $order->setExtensionAttributes($orderExtension);
        }
        return $order;
    }

    /**
     * @param $pickupInStoreId
     * @return string|null
     */
    protected function getShopCode($pickupInStoreId)
    {
        if (!$this->moduleManager->isEnabled('FiloBlu_Storelocator')) {
            return null;
        }

        if (empty($pickupInStoreId)) {
            return null;
        }

        $connection = $this->resourceConnection->getConnection();
        $table = $connection->getTableName('filoblu_storelocator_stores');
        return $connection->fetchOne(
            "SELECT store_code FROM $table WHERE store_id = ? LIMIT 1",
            [$pickupInStoreId]);
    }

    /**
     * @param $data
     * @return mixed
     */
    public function unserialize($data)
    {
        if (interface_exists(SerializerInterface::class) || class_exists(SerializerInterface::class)) {
            return ObjectManager::getInstance()->create(SerializerInterface::class)
                ->unserialize($data);
        }

        return unserialize($data);
    }

    /**
     * @param \Magento\Sales\Api\Data\OrderInterface $order
     * @return \Magento\Sales\Api\Data\OrderInterface|void
     */
    public function addGiftWrapping(OrderInterface $order)
    {
        if(!$this->moduleManager->isEnabled('Magento_GiftWrapping')) {
            return $order;
        }

        try {
            $isGiftWrappingAvailableForOrder = $this->scopeConfig->getValue(self::XML_PATH_ALLOWED_FOR_ORDER, ScopeInterface::SCOPE_STORE, $this->_storeManager->getStore());
            if (!$isGiftWrappingAvailableForOrder) {
                throw new Exception(
                    __('There is no order with provided id or gift wrapping isn\'t allowed')
                );
            }

            $wrappingId = $order->getGwId();
            if (!$wrappingId) {
                throw new Exception(__('There is no item with provided id in the order'));
            }

            $giftWrapping = $this->objectManager->get(WrappingFactory::class)->create()->load($wrappingId);
        } catch (Exception $e) {
            return $order;
        }

        $extensionAttributes = $order->getExtensionAttributes();
        /** @var OrderExtension $orderExtension */
        $orderExtension = $extensionAttributes ?: $this->orderExtensionFactory->create();
        $orderExtension->setGwId($giftWrapping->getId());
        $orderExtension->setGwAllowGiftReceipt($order->getGwAllowGiftReceipt());
        $orderExtension->setGwAddCard($order->getGwAddCard());
        $orderExtension->setGwPrice($order->getGwPrice());
        $orderExtension->setGwBasePrice($order->getGwBasePrice());
        $orderExtension->setGwItemsPrice($order->getGwItemsPrice());
        $orderExtension->setGwItemsBasePrice($order->getGwItemsBasePrice());
        $orderExtension->setGwCardPrice($order->getGwCardPrice());
        $orderExtension->setGwCardBasePrice($order->getGwCardBasePrice());
        $orderExtension->setGwBaseTaxAmount($order->getGwBaseTaxAmount());
        $orderExtension->setGwTaxAmount($order->getGwTaxAmount());
        $orderExtension->setGwItemsBaseTaxAmount($order->getGwItemsBaseTaxAmount());
        $orderExtension->setGwItemsTaxAmount($order->getGwItemsTaxAmount());
        $orderExtension->setGwCardBaseTaxAmount($order->getGwCardBaseTaxAmount());
        $orderExtension->setGwCardTaxAmount($order->getGwCardTaxAmount());
        $orderExtension->setGwPriceInclTax($order->getGwPriceInclTax());
        $orderExtension->setGwBasePriceInclTax($order->getGwBasePriceInclTax());
        $orderExtension->setGwCardPriceInclTax($order->getGwCardPriceInclTax());
        $orderExtension->setGwCardBasePriceInclTax($order->getGwCardBasePriceInclTax());
        $orderExtension->setGwItemsPriceInclTax($order->getGwItemsPriceInclTax());
        $orderExtension->setGwItemsBasePriceInclTax($order->getGwItemsBasePriceInclTax());
        $order->setExtensionAttributes($orderExtension);
        return $order;
    }
}
